/*-------------------------------------------------------------------------+
| Copyright 2019 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.tooling.base.utils;

import static java.util.Objects.requireNonNull;
import static org.fortiss.tooling.base.model.layout.LayoutPackage.ILAYOUTED_MODEL_ELEMENT__LAYOUT_DATA;

import java.util.ArrayList;
import java.util.EventObject;
import java.util.List;

import org.eclipse.emf.common.command.CommandStackListener;
import org.eclipse.emf.common.notify.Notification;
import org.eclipse.emf.ecore.EAttribute;
import org.eclipse.emf.ecore.EObject;
import org.eclipse.emf.ecore.util.EContentAdapter;
import org.fortiss.tooling.base.model.layout.ILayoutData;
import org.fortiss.tooling.base.model.layout.ILayoutedModelElement;
import org.fortiss.tooling.base.model.layout.LayoutPackage;
import org.fortiss.tooling.kernel.extension.data.ITopLevelElement;
import org.fortiss.tooling.kernel.service.ICommandStackService;
import org.fortiss.tooling.kernel.service.IPersistencyService;

/**
 * A specific {@link EContentAdapter} used for layouted model elements and layout data. It records
 * whether changes to the layout of the model element occurred, but triggers only after a command
 * stack change occurred afterwards.
 * 
 * @author hoelzl
 */
public class LayoutModelElementAdapter extends EContentAdapter implements CommandStackListener {
	/** The layouted model element. */
	private final ILayoutedModelElement modelElement;
	/** List of relevant {@link Notification}s recorded during a change transaction. */
	private final List<Notification> notifications = new ArrayList<>();

	/** Constructor. */
	public LayoutModelElementAdapter(ILayoutedModelElement layouted) {
		this.modelElement = requireNonNull(layouted);
		modelElement.eAdapters().add(this);
		ITopLevelElement root =
				IPersistencyService.getInstance().getTopLevelElementFor(modelElement);
		ICommandStackService.getInstance().addCommandStackListener(root, this);
	}

	/** Removes this adapter from the underlying model element. */
	public void unregisterAdapter() {
		modelElement.eAdapters().remove(this);
		ITopLevelElement root =
				IPersistencyService.getInstance().getTopLevelElementFor(modelElement);
		ICommandStackService.getInstance().removeCommandStackListener(root, this);
	}

	/** {@inheritDoc} */
	@Override
	public void notifyChanged(Notification notification) {
		Object notifier = notification.getNotifier();
		if(notifier == modelElement) {
			Object feature = notification.getFeature();
			if(feature instanceof EAttribute) {
				notifications.add(notification);
				// no selfAdapt needed for attributes
				return;
			}
			if(feature == LayoutPackage.Literals.ILAYOUTED_MODEL_ELEMENT__LAYOUT_DATA) {
				// self adapt to layout data
				super.notifyChanged(notification);
				notifications.add(notification);
				return;
			}
		} else if(notifier instanceof ILayoutData) {
			// self adapt to layout data
			super.notifyChanged(notification);
			notifications.add(notification);
			return;
		}
		// prevent self-adapt for all other cases
	}

	/** Called when layout data of the underlying model element changed. */
	protected void layoutChanged() {
		// default does nothing
	}

	/** {@inheritDoc} */
	@Override
	protected void selfAdapt(Notification notification) {
		// ignore resources and resource sets
		if(notification.getNotifier() instanceof EObject) {
			super.selfAdapt(notification);
		}
	}

	/** {@inheritDoc} */
	@Override
	protected void handleContainment(Notification notification) {
		// only adapt to layout data
		if(notification.getNotifier() instanceof ILayoutData) {
			super.handleContainment(notification);
			return;
		}

		Class<?> instanceClass = LayoutPackage.Literals.ILAYOUTED_MODEL_ELEMENT.getInstanceClass();
		// only adapt to layout data
		if(notification.getFeatureID(instanceClass) == ILAYOUTED_MODEL_ELEMENT__LAYOUT_DATA) {
			super.handleContainment(notification);
		}
	}

	/** {@inheritDoc} */
	@Override
	public void commandStackChanged(EventObject event) {
		if(notifications.isEmpty()) {
			return;
		}
		layoutChanged();
		notifications.clear();
	}

	/** Returns list of collected notifications. */
	public List<Notification> getNotifications() {
		return notifications;
	}
}
